import os
import sys
import c4d
from c4d import bitmaps, plugins, NOTOK, utils, gui


UV2O_MASTERTAG_ID = 1023532
UV2O_TAG_ID = 1023530
UV2O_ID = 1012108
UV2O_SETTINGS_ID = 1023531

_ids_ = dict( #write it like this so i can just copy paste stuff from c4d_symbols
  IDS_UVTOOBJECT              = 100000,
  IDS_UVTOOBJECT_HELP         = 100002,
  IDS_UVCOPYTAG               = 100001,
  IDS_UVCOPYTAG_HELP          = 100016,
  IDS_UVTOOBJECT_MASTER       = 100003,
  IDS_UVTOOBJECT_MASTER_HELP  = 100015,

  IDS_UVTOOBJECT_NAME_BG      = 100004,
  IDS_UVTOOBJECT_NAME_CAM     = 100005,

  IDS_UVTOOBJECT_DLG_TITLE      = 100006,
  IDS_UVTOOBJECT_DLG_SCENE      = 100007,
  IDS_UVTOOBJECT_DLG_SIZE_W     = 100008,
  IDS_UVTOOBJECT_DLG_SIZE_H     = 100009,
  IDS_UVTOOBJECT_DLG_HELP       = 100017,
  
  IDS_UVTOOBJECT_NAME_UVGROUP  = 100010,
  IDS_UVTOOBJECT_NAME_UVFRAME  = 100011,
  IDS_UVTOOBJECT_NAME_OBJGROUP = 100012,

  IDS_EQUALIZE_AREAS           = 100013,
  IDS_EQUALIZE_AREAS_HELP      = 100014,

  DLG_UVTOOBJECT       = 9999,

  ID_UVTOOBJECT_SCENE    = 3000,
  ID_UVTOOBJECT_SIZE_W   = 4000,
  ID_UVTOOBJECT_SIZE_H   = 5000,
  ID_UVTOOBJECT_DLG_HELP = 6000,
)
globals().update(_ids_)


def FindUVWTag(obj):
  tags = obj.GetTags()
  firstTag = None
  firstActive = None
  for t in tags:
    if t.CheckType(c4d.Tuvw):
      if not firstTag:
        firstTag = t
      if not firstActive and t.GetBit(c4d.BIT_ACTIVE):
        firstActive = t
  if firstActive:
    return firstActive
  else:
    return firstTag


def FindUVFrame(parent):
  c = parent.GetDown()
  while c:
    if c.CheckType(c4d.Osplinerectangle):
      return c
    c = c.GetNext()
  return None


def AddUVWObjects(parent, frame, doc, objectlist):
  objectlist = map(lambda o: (o, FindUVWTag(o)), objectlist)
  objectlist = filter(lambda (o,t): t <> None, objectlist)
  for obj, uvwtag in objectlist:
    uvwobj = c4d.PolygonObject(0, 0)
    if frame and frame.GetPred():
      uvwobj.InsertBefore(frame)
    else:
      uvwobj.InsertUnderLast(parent)
    doc.AddUndo(c4d.UNDOTYPE_NEW, uvwobj)
    uvwobj.SetName(obj.GetName() + '-UV')
    copytag = uvwobj.MakeTag(UV2O_TAG_ID)
    copytag[c4d.UVC_TARGET_UV] = uvwtag
    copytag[c4d.UVC_FRAME]     = frame
    msg = c4d.BaseContainer()
    msg.SetId(c4d.UVC_REBUILD)
    copytag.Message(c4d.MSG_BASECONTAINER, msg)


def AddCamera(parent, frame, doc):
  w = frame[c4d.PRIM_RECTANGLE_WIDTH]
  h = frame[c4d.PRIM_RECTANGLE_HEIGHT]
  cam = c4d.BaseObject(c4d.Ocamera)
  #cam.SetMl(utils.MatrixMove(c4d.Vector(w*0.5, h*0.5, -max(w, h)*0.5)))
  cam.SetMl(utils.MatrixMove(c4d.Vector(0, 0, -max(w, h)*0.5)))
  cam.SetName(c4d.plugins.GeLoadString(IDS_UVTOOBJECT_NAME_CAM))
  cam[c4d.CAMERAOBJECT_APERTURE] = 2.0
  cam[c4d.CAMERA_FOCUS] = 1.0
  cam.InsertUnderLast(frame)
  doc.AddUndo(c4d.UNDOTYPE_NEW, cam)


def AddBackgroundPlane(parent, frame, doc):
  w = frame[c4d.PRIM_RECTANGLE_WIDTH]
  h = frame[c4d.PRIM_RECTANGLE_HEIGHT]
  obj = c4d.BaseObject(c4d.Oplane)
  #obj.SetMl(utils.MatrixMove(c4d.Vector(w*0.5, h*0.5, 1.0)))
  obj.SetMl(utils.MatrixMove(c4d.Vector(0, 0, 1)))
  obj.SetName(c4d.plugins.GeLoadString(IDS_UVTOOBJECT_NAME_BG))        
  obj[c4d.PRIM_PLANE_WIDTH] = w
  obj[c4d.PRIM_PLANE_HEIGHT] = h
  obj[c4d.PRIM_PLANE_SUBW] = 1
  obj[c4d.PRIM_PLANE_SUBH] = 1
  obj[c4d.PRIM_AXIS] = c4d.PRIM_AXIS_ZN
  t = obj.MakeTag(c4d.Tdisplay)
  t[c4d.DISPLAYTAG_SDISPLAYMODE] = c4d.DISPLAYTAG_SDISPLAY_FLAT_WIRE
  t[c4d.DISPLAYTAG_AFFECT_DISPLAYMODE] = True
  obj.InsertUnderLast(frame)
  doc.AddUndo(c4d.UNDOTYPE_NEW, obj)

  
def AdjustRenderData(parent, frame, doc):
  rd = doc.GetActiveRenderData()
  w = frame[c4d.PRIM_RECTANGLE_WIDTH]
  h = frame[c4d.PRIM_RECTANGLE_HEIGHT]
  if h > 0 and w > 0:
    doc.AddUndo(c4d.UNDOTYPE_CHANGE, rd)
    rd[c4d.RDATA_PIXELPRESET] = c4d.RDATA_PIXELPRESET_SQUARE
    rd[c4d.RDATA_PIXELASPECT] = 1
    rd[c4d.RDATA_FILMASPECT] = w/h
    rd[c4d.RDATA_LOCKRATIO] = True
    rd[c4d.RDATA_YRES] = rd[c4d.RDATA_XRES] * h/w
    rd.Message(c4d.MSG_UPDATE)
  

def CreateUVToObjectThings(doc, prefs):
  makeNewScene = prefs[ID_UVTOOBJECT_SCENE]
  uvarea_width = prefs[ID_UVTOOBJECT_SIZE_W]
  uvarea_height = prefs[ID_UVTOOBJECT_SIZE_H]

  objectlist = doc.GetActiveObjectsFilter(True, NOTOK, c4d.Opolygon)
    
  if makeNewScene:
    newdoc = c4d.documents.BaseDocument()
    objgroup = c4d.BaseObject(c4d.Onull)
    objgroup.SetName(c4d.plugins.GeLoadString(IDS_UVTOOBJECT_NAME_OBJGROUP))
    newdoc.InsertObject(objgroup)
    clones = []
    for obj in objectlist:
      clone = obj.GetClone(c4d.COPYFLAGS_NO_HIERARCHY | c4d.COPYFLAGS_NO_BRANCHES)
      clone.InsertUnderLast(objgroup)
      for t in obj.GetTags():
        if any(t.CheckType(x) for x in (c4d.Tphong, c4d.Tpolygon, c4d.Tpoint, c4d.Tuvw)):
          tclone = t.GetClone(c4d.COPYFLAGS_0)
          tclone.SetAllBits(t.GetAllBits())
          clone.InsertTag(tclone)
      clones.append(clone)
    objectlist = clones
    
    newdoc.SetDocumentName(doc.GetDocumentName()+'-UV')
    c4d.documents.InsertBaseDocument(newdoc)
    c4d.documents.SetActiveDocument(newdoc)
    doc = newdoc
    
  doc.StartUndo()
    
  parent = c4d.BaseObject(c4d.Onull)
  parent.SetName(c4d.plugins.GeLoadString(IDS_UVTOOBJECT_NAME_UVGROUP))
  parent.MakeTag(UV2O_MASTERTAG_ID)
  doc.InsertObject(parent)
  doc.AddUndo(c4d.UNDOTYPE_NEW, parent)
  
  frame = c4d.BaseObject(c4d.Osplinerectangle)
  frame[c4d.PRIM_RECTANGLE_WIDTH] = uvarea_width
  frame[c4d.PRIM_RECTANGLE_HEIGHT] = uvarea_height
  frame.SetMl(utils.MatrixMove(c4d.Vector(uvarea_width*0.5, uvarea_height*0.5, 0)))
  frame.SetName(c4d.plugins.GeLoadString(IDS_UVTOOBJECT_NAME_UVFRAME))
  frame.InsertUnderLast(parent)
  
  AddUVWObjects(parent, frame, doc, objectlist)

  if makeNewScene:
      AdjustRenderData(parent, frame, doc)
  
  doc.EndUndo()
  c4d.EventAdd()


  
class UV2ObjectMasterTag(plugins.TagData):
  def Message(self, tag, msg, data):
    if msg == c4d.MSG_DESCRIPTION_COMMAND:
      parent = tag.GetObject()
      frame  = FindUVFrame(parent)
      doc    = tag.GetDocument()
    
      if data['id'][0].id == c4d.UV2OBJECT_MASTER_ADD_OBJECT:       
        objectlist = doc.GetActiveObjectsFilter(True, NOTOK, c4d.Opolygon)
        doc.StartUndo()
        AddUVWObjects(parent, frame, doc, objectlist)
        doc.EndUndo()
        c4d.EventAdd()

      if data['id'][0].id == c4d.UV2OBJECT_MASTER_CREATE_CAMERA:
        doc.StartUndo()        
        AddCamera(parent, frame, doc)
        doc.EndUndo()
        c4d.EventAdd()
       
      if data['id'][0].id == c4d.UV2OBJECT_MASTER_CREATE_BACKGROUND:
        doc.StartUndo()
        AddBackgroundPlane(parent, frame, doc)
        doc.EndUndo()
        c4d.EventAdd()
      
      if data['id'][0].id == c4d.UV2OBJECT_MASTER_ADJUST_RENDERDATA:
        doc.StartUndo()
        AdjustRenderData(parent, frame, doc)
        doc.EndUndo()
        c4d.EventAdd()
    return True

    
def GetPrefs(doc):
  bc = c4d.BaseContainer()
  bc[ID_UVTOOBJECT_SIZE_W] = 100
  bc[ID_UVTOOBJECT_SIZE_H] = 100
  bc[ID_UVTOOBJECT_SCENE]  = False
  wp = c4d.plugins.GetWorldPluginData(UV2O_SETTINGS_ID)
  if wp:
    bc.MergeContainer(wp)
  return bc

    
def SetPrefs(doc, bc):
  c4d.plugins.SetWorldPluginData(UV2O_SETTINGS_ID, bc, add = False)
  

class UVToObjectDialog(gui.GeDialog):
    def Set(self, fun, id, **kwargs):
      getattr(self, fun)(id, self.bc[id], **kwargs)
    
    def Get(self, fun, id):
      self.bc[id] = getattr(self, fun)(id)
    
    def CreateLayout(self):
      self.result = False
      return gui.GeDialog.CreateLayout(self) and self.LoadDialogResource(DLG_UVTOOBJECT)
      
    def InitValues(self):
      self.Set('SetReal', ID_UVTOOBJECT_SIZE_W, min = 0, step = 10, format = c4d.FORMAT_METER)
      self.Set('SetReal', ID_UVTOOBJECT_SIZE_H, min = 0, step = 10, format = c4d.FORMAT_METER)
      self.Set('SetBool', ID_UVTOOBJECT_SCENE);
      return gui.GeDialog.InitValues(self)
      
    def Command(self, id, bc):
      self.Get('GetReal', ID_UVTOOBJECT_SIZE_W)
      self.Get('GetReal', ID_UVTOOBJECT_SIZE_H)
      self.Get('GetBool', ID_UVTOOBJECT_SCENE)
      if id in (c4d.IDC_OK, c4d.IDC_CANCEL):
        self.result = True if id == c4d.IDC_OK else False
        self.Close()
      if id == ID_UVTOOBJECT_DLG_HELP:
        OpenHTML(os.path.join(os.path.dirname(__file__), 'documentation', 'uv2object.html'))
      return True

      
class UVToObject(c4d.plugins.CommandData):
    def __init__(self):
      self.dialog = UVToObjectDialog()

    def Execute(self, doc):
      prefs = GetPrefs(doc)
      CreateUVToObjectThings(doc, prefs)
      return True

    def RestoreLayout(self, sec_ref):
      return self.dialog.Restore(pluginid=UV2O_ID, secret=sec_ref)

    def ExecuteOptionID(self, doc, plugid, subid):
      prefs = GetPrefs(doc)
      self.dialog.doc = doc
      self.dialog.bc  = prefs
      self.dialog.Open(c4d.DLG_TYPE_MODAL)
      SetPrefs(doc, prefs)
      if self.dialog.result:
        CreateUVToObjectThings(doc, prefs)
      return True


def OpenHTML(url):
  import subprocess
  #shamelessly copy pasted
  # http://stackoverflow.com/questions/4216985/call-to-operating-system-to-open-url
  if sys.platform=='win32':
      os.startfile(url)
  elif sys.platform=='darwin':
      subprocess.Popen(['open', url])
  else:
      try:
          subprocess.Popen(['xdg-open', url])
      except OSError:
          print 'Please open a browser on: '+url


if __name__ == "__main__":
    bmp = bitmaps.BaseBitmap()
    dir, file = os.path.split(__file__)
    fn = os.path.join(dir, "res", "uvtoobject.png")
    bmp.InitWith(fn)
    plugins.RegisterTagPlugin(id=UV2O_MASTERTAG_ID, 
                              str=c4d.plugins.GeLoadString(IDS_UVTOOBJECT_MASTER),
                              info=c4d.TAG_EXPRESSION|c4d.TAG_VISIBLE, g=UV2ObjectMasterTag,
                              description="tuvtoobjectmaster", icon = bmp)

    c4d.plugins.RegisterCommandPlugin(id=UV2O_ID, 
                                      str=c4d.plugins.GeLoadString(IDS_UVTOOBJECT),
                                      help=c4d.plugins.GeLoadString(IDS_UVTOOBJECT_HELP),
                                      info=c4d.PLUGINFLAG_COMMAND_OPTION_DIALOG,
                                      dat=UVToObject(), icon=bmp)
